/**
 * \file Mini.cpp
 * \brief Contains code for methods listed in Mini.hh
 * \author <a href="mailto:auto_pilot@gmx.at">Hoermann Klaus</a>
 * \author <a target="_blank" href="http://www.netmusic.at.tt">www.netmusic.at.tt</a>
 * \version 0.1
 * \date 24.03.2004
 * \date last modified 03.06.2005
 *
 */

/**
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation; either version 2.1 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 **/
 
#include <cstring>
#include <cmath>
#include <cstdio>
#include <iostream>

#include "Mini.hh"
#include "Netmusic.hh"

//****************************************************************
void Mini::set_mini_word(int* noteArray, int k, char* binaryMiniWord,
	int noteOffValue, int* size_mem_mpz, int startKey )
{
	mpz_t tempWord1, tempWord2, miniWord;
	char* tmp;

	//init mpz
	mpz_init(miniWord);
	mpz_init(tempWord1);
	mpz_init(tempWord2);
	mpz_set_ui(miniWord,0);

	// mini calculations
	Mini::sort_array(noteArray, k);
	Mini::ksub_rank( k, noteArray, miniWord);
	mpz_set(tempWord1, miniWord);
	mpz_set(tempWord2, miniWord);
	*size_mem_mpz = Mini::get_size_mem_mpz(tempWord1);
	tmp = new char[(*size_mem_mpz)];
	Mini::mpz_2_bin(tempWord2, k, noteOffValue, tmp);
	
	// set pointer
	binaryMiniWord = tmp;

	// print
	std::cout << DELIMITER << std::endl;
	Mini::print_array(noteArray, k, startKey);// print array
	gmp_printf("\n rank: %Zd\n", miniWord);
	Mini::print_mem( binaryMiniWord, (*size_mem_mpz));
	std::cout << " mpz_size: " << (*size_mem_mpz) << std::endl;
	std::cout << " k:        " << k << std::endl;
	std::cout << " noteOff:  " << noteOffValue << std::endl << std::endl;
}

//***************************************************************
void Mini::ksub_rank ( int k, int* a, mpz_t rank )
{
  int i, j;
	int first = 0;
  mpz_t iprod;
  mpz_init(iprod);

	// check if it is the first chord [0], [0][1], [0][1][2]
	for(i=1;i<=k;i++)
	{
		// check if it is first arrray
		if(a[i-1] == (i-1))
			first = 1;
		// check if array contains a relative note smaller 0
		else if(a[i-1] < 0)
		{
			mpz_set_ui(rank,0);
			std::cout << "Played Note smaller than start Key!" << std::endl;
			return;
		}
		else
		{
			first = 0;
			break;
		}
	}

	// set rank to 0
	if(first==1)
		mpz_set_ui(rank,0);

	// calculate rank
	else
	{
		// sum up to rank
		for(i=1;i<=k;i++)
		{
			mpz_set_ui(iprod,1);

			for (j=i+1; j<=a[i-1]-1; j++)
				mpz_mul_ui(iprod,iprod,j);

			for (j=1; j<=a[i-1]-i-1; j++)
				mpz_cdiv_q_ui(iprod,iprod,j);

			if(a[i-1]==i) /* bug: a[i-1] == 1, modified by GR 14.05.04 */
				mpz_set_ui(iprod,0);

			mpz_add(rank, rank, iprod);
		}

		mpz_add_ui(rank, rank, 1);
		mpz_clear(iprod);
	}
  return;
}

//***************************************************************
void Mini::ksub_unrank ( int k, int* a, mpz_t rank )
{
	int i, ip;
  mpz_t jrank, iprod, tmp1;

  mpz_init(jrank);
  mpz_init(iprod);
  mpz_init(tmp1);

	// rank is 0
	if(mpz_cmp_ui(rank,0)==0)
	{
		for(i=0;i<k;i++)
		{
			a[i] = i;
		}
	}
	// rank is not 0
	else
	{
		mpz_sub_ui(jrank, rank, 1);

		for(i=k; 1<=i; i--)
		{
			ip=i-1;
			mpz_set_ui(iprod,1);

			for(;;)
			{
				ip++;
				if(ip!=i)
				{
					mpz_mul_ui(tmp1,iprod, ip);
					mpz_cdiv_q_ui(iprod, tmp1, (ip-i));
				}
				if(mpz_cmp(jrank,iprod)<0)
					break;
			}

			if(ip!=i)
			{
				mpz_mul_ui(tmp1, iprod, (ip-i));
				mpz_cdiv_q_ui(iprod, tmp1, ip);
			}

			mpz_sub(jrank, jrank, iprod);
			a[i-1]=ip;
		}

		mpz_clear(jrank);
		mpz_clear(iprod);
		mpz_clear(tmp1);
	}
  return;
}

//*****************************************************************
int Mini::get_bin_digits_mpz(mpz_t number)
{
	int count=1;
	if( mpz_cmp_ui(number,1) == 0 )
		return 1;

	mpz_fdiv_q_ui(number,number,2);
	while( (mpz_cmp_ui(number,0)) > 0 )
	{
		count++;
		if( mpz_cmp_ui(number,1) == 0)
			return count;
		mpz_fdiv_q_ui(number,number,2);
	}
	return count;
}

//*****************************************************************
int Mini::get_bin_digits_int(int number)
{
	int count=1;
	if(number==1)
		return 1;
	while( (number=number/2)>0 )
		count++;
	return count;
}

//***************************************************************
void Mini::mpz_2_bin(mpz_t value, int k, int note, char* mem_pointer)
{
	int digits_value, digits_k, digits_size, size_mem;
	char* location; // pointer for going thru the bin_val
	
	mpz_t value_temp, value_temp2, remainder;
	mpz_init(value_temp);
	mpz_init(value_temp2);
	mpz_init(remainder);
	mpz_set(value_temp, value);
	mpz_set(value_temp2, value);
	mpz_set_ui(remainder,0);

	// calculate the amount of necessary digits
	digits_k = get_bin_digits_int(k);
	digits_value = get_bin_digits_mpz(value);

	// caclulate size of buffer in memory
	size_mem = get_size_mem_mpz(value_temp2);
	digits_size = get_bin_digits_int(size_mem);

	int k_tmp = k;
	int size_mem_tmp = size_mem;

	// set to start of lowest byte
	location = mem_pointer + (size_mem-1);

	//-------------------------------------------------------------
	// shift CHORD CODE in memory
	for(int i=0;i<((size_mem-1)*8); i++)
	{
		// fill value
		if(i<digits_value)
		{
			mpz_mod_ui(remainder, value_temp, 2); // mod 2
			if( mpz_cmp_ui(remainder,1) == 0  ) // compare remainder with 1
			{
				*location = (*location) >> 1;
				*location = (*location) | 128;
			}
			else if( mpz_cmp_ui(remainder,0) == 0 ) // compare remainder with 0
			{
				*location = (*location) >> 1;
				*location = (*location) & 127;
			}
			mpz_fdiv_q_ui(value_temp,value_temp,2);
		}
		// fill zeros
		else
		{
				*location = (*location) >> 1;
				*location = (*location) & 127;
		}

		// switch to another char in memory
		if(((i+1)%8)==0)
			location--;
	}

	//-------------------------------------------------------------
	// shift VOICE CODE in memory
	for(int i=0;i<DIGITS_K_DEF;i++)
	{
		// fill value
		if(i<digits_k)
		{
			if(k_tmp%2==1)
			{
				*location = (*location) >> 1; // leftshift with 0
				*location = (*location) | 128; // set to 1
			}
			else
			{
				*location = (*location) >> 1;
				*location = (*location) & 127; // set to 0
			}
			k_tmp=k_tmp/2;
		}
		// fill zeros
		else
		{
				*location = (*location) >> 1;
				*location = (*location) & 127; // set to 0
		}
	}

	//-------------------------------------------------------------
	// shift in SIZE CODE
	for(int i=0;i<DIGITS_SIZE_DEF;i++)
	{
		// fill value
		if(i<digits_size)
		{
			if((size_mem_tmp%2)==1)
			{
				*location = (*location) >> 1;
				*location = (*location) | 128;
			}
			else
			{
				*location = (*location) >> 1;
				*location = (*location) & 127;
			}
			size_mem_tmp = size_mem_tmp/2;
		}
		// fill zeros
		else
		{
				*location = (*location) >> 1;
				*location = (*location) & 127;
		}
	}

	//------------------------------------------------------------
	// shift NOTE ON/OFF command in memory
	if(note == 1)
	{
	  *location = (*location) >> 1;
	  *location = (*location) | 128;
	}
	else if(note == 0)
	{
	  *location = (*location) >> 1;
	  *location = (*location) & 127;
	}
}

//***************************************************************
void Mini::bin_2_mpz ( char* mem_pointer, int* k, mpz_t value, int* note)
{
	mpz_set_ui(value,0);
	*k = 0;
	*note = 0;

	char first = mem_pointer[0];

	// read note on/off
	if(first & 128)
		*note = 1;
	else
		*note = 0;

	// read size
	int size_mem = 0;
	if(first & 64)
		size_mem += 4;
	if(first & 32)
		size_mem += 2;
	if(first & 16)
		size_mem += 1;

	// read k
	if(first & 8)
		*k += 8;
	if(first & 4)
		*k += 4;
	if(first & 2)
		*k += 2;
	if(first & 1)
		*k += 1;

	// read value
	for(int i=0; i<size_mem-1; i++)
	{
		char temp = mem_pointer[size_mem-1-i];

		if(temp & 1)
			mpz_add_ui(value, value, (int)pow(2.0, (double)(0+(i*8))));
		if(temp & 2)
			mpz_add_ui(value, value, (int)pow(2.0, (double)(1+(i*8))));
		if(temp & 4)
			mpz_add_ui(value, value, (int)pow(2.0, (double)(2+(i*8))));
		if(temp & 8)
			mpz_add_ui(value, value, (int)pow(2.0, (double)(3+(i*8))));
		if(temp & 16)
			mpz_add_ui(value, value, (int)pow(2.0, (double)(4+(i*8))));
		if(temp & 32)
			mpz_add_ui(value, value, (int)pow(2.0, (double)(5+(i*8))));
		if(temp & 64)
			mpz_add_ui(value, value, (int)pow(2.0, (double)(6+(i*8))));
		if(temp & 128)
			mpz_add_ui(value, value, (int)pow(2.0, (double)(7+(i*8))));
	}
}

//*****************************************************************
void Mini::sort_array( int* noteArray, int size )
{
	int pass=1, pair=1, temp=0;

	for(pass=1;pass<size;pass++)
	{
		for(pair=1;pair<size;pair++)
		{
			if(noteArray[pair-1]>noteArray[pair])
			{
				temp = noteArray[pair-1];
				noteArray[pair-1]=noteArray[pair];
				noteArray[pair]=temp;
			}
		}
	}
}

//*****************************************************************
void Mini::print_array( int* noteArray, int size, int startKey )
{
	for(int i=0;i<size;i++)
	{
		std::cout << " noteArray[" << i << "]: " << noteArray[i] << 
			" + startKey(" << startKey << ") = " 
			<< noteArray[i]+startKey << std::endl;
	}
}

//*****************************************************************
void Mini::extract_size(char firstRead, int* size)
{
	char tmp = firstRead;
	*size = 0;
	
	// size		
	if(tmp & 64)
		(*size) += 4;
	if(tmp & 32)
		(*size) += 2;
	if(tmp & 16)
		(*size) += 1;	
}

//*****************************************************************
void Mini::print_mem(char* mem_pointer, int size_bin)
{
	std::cout << " content:      ";
	for(int i=0;i<size_bin;i++)
	{
		char b = mem_pointer[i];
		std::cout << "{ ";
		for (int j=0; j<8; j++)
		{
			if(b & 128)
				std::cout << "1";
			else
				std::cout << "0";

			b = b << 1; // um eine stelle nach links shiften

			// separate note on/off
			if(i==0 && j==0)
				std::cout << " | ";
			// separate size
			if(i==0 && j==3)
				std::cout << " | ";
		}
		std::cout << " } ";
	}
	std::cout << std::endl;
}
